import 'dart:async';
import 'dart:math';

import 'package:equatable/equatable.dart';
import 'package:flutter/material.dart';
import 'package:hydrated_bloc/hydrated_bloc.dart';
import 'package:sudoku/config/default_global_settings.dart';
import 'package:sudoku/models/cell.dart';

import 'package:sudoku/models/cell_location.dart';
import 'package:sudoku/models/game.dart';
import 'package:sudoku/models/settings_game.dart';
import 'package:sudoku/models/settings_global.dart';
import 'package:sudoku/utils/board_animate.dart';

part 'game_state.dart';

class GameCubit extends HydratedCubit<GameState> {
  GameCubit()
      : super(GameState(
          game: Game.createNull(),
        ));

  void updateState(Game game) {
    emit(GameState(
      game: game,
    ));
  }

  void refresh() {
    updateState(Game(
      gameSettings: state.game.gameSettings,
      globalSettings: state.game.globalSettings,
      board: state.game.board,
      solvedBoard: state.game.solvedBoard,
      isRunning: state.game.isRunning,
      isStarted: state.game.isStarted,
      isFinished: state.game.isFinished,
      blockSizeHorizontal: state.game.blockSizeHorizontal,
      blockSizeVertical: state.game.blockSizeVertical,
      boardSize: state.game.boardSize,
      shuffledCellValues: state.game.shuffledCellValues,
      boardConflicts: state.game.boardConflicts,
      selectedCell: state.game.selectedCell,
      showConflicts: state.game.showConflicts,
      givenTipsCount: state.game.givenTipsCount,
      buttonTipsCountdown: state.game.buttonTipsCountdown,
      animationInProgress: state.game.animationInProgress,
      boardAnimated: state.game.boardAnimated,
    ));
  }

  void startNewGame({
    required GameSettings gameSettings,
    required GlobalSettings globalSettings,
  }) {
    final Game newGame = Game.createNew(
      gameSettings: gameSettings,
      globalSettings: globalSettings,
    );

    newGame.dump();

    updateState(newGame);
    refresh();

    BoardAnimate.startAnimation(this, 'start');
  }

  void selectCell(CellLocation location) {
    state.game.selectedCell = state.game.board.get(location);
    refresh();
  }

  void unselectCell() {
    state.game.selectedCell = null;
    refresh();
  }

  void updateCellValue(CellLocation location, int value) {
    if (state.game.board.get(location).isFixed == false) {
      state.game.board.set(
        location,
        Cell(
          location: location,
          value: value,
          isFixed: false,
        ),
      );
      state.game.isStarted = true;
      refresh();
    }

    if (state.game.checkBoardIsSolved()) {
      BoardAnimate.startAnimation(this, 'win');
      state.game.isFinished = true;
      refresh();
    }
  }

  void toggleShowConflicts() {
    state.game.showConflicts = !state.game.showConflicts;
    refresh();
  }

  void increaseGivenTipsCount() {
    state.game.givenTipsCount++;
    state.game.buttonTipsCountdown = DefaultGlobalSettings.defaultTipCountDownValueInSeconds;
    refresh();

    const Duration interval = Duration(milliseconds: 500);
    Timer.periodic(
      interval,
      (Timer timer) {
        if (state.game.buttonTipsCountdown == 0) {
          timer.cancel();
        } else {
          state.game.buttonTipsCountdown = max(state.game.buttonTipsCountdown - 1, 0);
        }
        refresh();
      },
    );
  }

  void quitGame() {
    state.game.isRunning = false;
    refresh();
  }

  void resumeSavedGame() {
    state.game.isRunning = true;
    refresh();
  }

  void deleteSavedGame() {
    state.game.isRunning = false;
    state.game.isFinished = true;
    refresh();
  }

  void updateAnimationInProgress(bool animationInProgress) {
    state.game.animationInProgress = animationInProgress;
    refresh();
  }

  void setAnimatedBackground(List animatedCellsPattern) {
    for (int row = 0; row < state.game.boardSize; row++) {
      for (int col = 0; col < state.game.boardSize; col++) {
        state.game.boardAnimated[row][col] = animatedCellsPattern[row][col];
      }
    }
    refresh();
  }

  void resetAnimatedBackground() {
    for (int row = 0; row < state.game.boardSize; row++) {
      for (int col = 0; col < state.game.boardSize; col++) {
        state.game.boardAnimated[row][col] = false;
      }
    }
    refresh();
  }

  @override
  GameState? fromJson(Map<String, dynamic> json) {
    Game game = json['game'] as Game;

    return GameState(
      game: game,
    );
  }

  @override
  Map<String, dynamic>? toJson(GameState state) {
    return <String, dynamic>{
      'game': state.game.toJson(),
    };
  }
}
